/*
    DFT++ is a density functional package developed by the research group
    of Professor Tomas Arias

    Copyright 1996-2003 Sohrab Ismail-Beigi

    This file is part of DFT++.

    DFT++ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DFT++ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DFT++; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please see the file CREDITS for a list of authors.

    For academic users, we request that publications using results obtained with
    this software reference

    "New algebraic formulation of density functional calculation," by Sohrab Ismail-Beigi
    and T.A. Arias, Computer Physics Communications 128:1-2, 1-45 (June 2000).

    and, if using the wavelet basis, further reference

    "Multiresolution analysis of electronic structure: semicardinal and wavelet bases,"
    T.A. Arias, Reviews of Modern Physics 71:1, 267-311 (January 1999).

    and 

    "Robust ab initio calculation of condensed matter: transparent convergence through
    semicardinal multiresolution analysis,'' I.P. Daykov, T.A. Arias, and
    Torkel D. Engeness, Physical Review Letters, 90:21, 216402 (May 2003).

    For your convenience, preprints of the above articles may be obtained from
    http://arXiv.org/abs/cond-mat/9909130, 9805262, and 0204411, respectively.
*/

/*---------------------------  Basis  -----------------------------------*
 *                                                                       *
 * Master structure defining the basis set used for the calculation      *
 *  One Basis object each is needed for every k-point.                   *
 *  And an additional Basis object is used for charge-density-FFT.       *
 *                                                                       *
 * (operated on in setup.c , setup_basis.c)                              *
 *                                                                       *
 *-----------------------------------------------------------------------*/

#ifndef DFT_PW_BASIS_H
#define DFT_PW_BASIS_H


class PW_BasisSpec
{
 public:
  // Plane-wave cutoff energy in Hartrees
  real Ecut;
  
  Lattice *lattice;

  // sizes for real-space grid and their product (i.e. FFT box size)
  // this is used to store the initial values, if they are given
  int size_realgrid;
  
  int Nx,Ny,Nz,NxNyNz;

  // Whether fft box size is computed manually or automatically
  int auto_fftbox;

  // What kind of basis we have (read from parser, passed to setup)
  int basis_flag; // 0 = centered at gamma
                  // 1 = kpoint dependent

  void setup(Everything &e);

  
  int kpt_fold[3];

  PW_BasisSpec();

};

class PW_Basis
{
 public:

  PW_BasisSpec *basis_spec; // points to global BasisSpec
  QuantumNumber *qnum;

  int nbasis; /* number of basis elements (i.e. G-vectors) */

  /* Minimum and maximum values of each G-direction (inclusive) */
  int Gxmin,Gxmax,Gymin,Gymax,Gzmin,Gzmax;


  int *Gx,*Gy,*Gz; /* the G-vectors for the basis in recip. lattice coords */
  int *index; /* indices of the basis functions in the FFT boxes used */

  // These are auxiliary column-sized arrays to store precomputed
  // values. eg. they are used in PHLO operators.
  real *kplusG2;
  real *kplusGx,*kplusGy,*kplusGz;
  real *invLkernel;
 
  PW_Basis();
  ~PW_Basis();

  void operator=(PW_Basis &);

  // Sets up the real-space grid
  void setup_template(PW_BasisSpec &);
  // Figures out the G vectors for a given k vector
  void setup(const vector3 &kvec);

  void init(){printf("Initializing PW_Basis\n");}

};

#endif // DFT_PW_BASIS_H
