/*
    DFT++ is a density functional package developed by the research group
    of Professor Tomas Arias

    Copyright 1996-2003 Sohrab Ismail-Beigi

    This file is part of DFT++.

    DFT++ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DFT++ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DFT++; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please see the file CREDITS for a list of authors.

    For academic users, we request that publications using results obtained with
    this software reference

    "New algebraic formulation of density functional calculation," by Sohrab Ismail-Beigi
    and T.A. Arias, Computer Physics Communications 128:1-2, 1-45 (June 2000).

    and, if using the wavelet basis, further reference

    "Multiresolution analysis of electronic structure: semicardinal and wavelet bases,"
    T.A. Arias, Reviews of Modern Physics 71:1, 267-311 (January 1999).

    and 

    "Robust ab initio calculation of condensed matter: transparent convergence through
    semicardinal multiresolution analysis,'' I.P. Daykov, T.A. Arias, and
    Torkel D. Engeness, Physical Review Letters, 90:21, 216402 (May 2003).

    For your convenience, preprints of the above articles may be obtained from
    http://arXiv.org/abs/cond-mat/9909130, 9805262, and 0204411, respectively.
*/

#include "header.h"

Speciesinfo::Speciesinfo(double Z_, char name_[], int natoms_, vector3* atpos_, double mass_, int* move_,
			 vector3* forces_, vector3* old_forces_, vector3* ion_vel_, double ion_damp_, double ion_fac_):
  Z(Z_),natoms(natoms_),atpos(atpos_),mass(mass_),move(move_),forces(forces_),old_forces(old_forces_),ion_vel(ion_vel_),
  ion_damp(ion_damp_),ion_fac(ion_fac_)
{
  strcpy(name, name_);
}

Speciesinfo::~Speciesinfo()
{
  if (natoms>0)
    {
      myfree(atpos);
      myfree(forces);
      myfree(old_forces);
      myfree(move);
      myfree(ion_vel);
    }
}

void Speciesinfo::setup(Everything &e)
{
  // If we have a species with no atoms, that is bad.  Complain and die
  if (natoms == 0)
    die("\n\nSpecies '%s' was defined but there are no such atoms.\nPlease enter some such atoms, or remove the species definition.\n\n",name);

  // some pots might need this
  pot.Z = Z;

  pot.setup(e);
}

void Speciesinfo::print(const Output *out, Everything &e)
{
  vector3 xx;
  out->printf(">> Species %s:\n",name);
  out->printf(">> Z = %f\tmass = %f\tnatoms = %d\n>> Positions follow in LATTICE coordinates:\n",
	    Z, mass, natoms);
  for (int i=0; i < natoms; i++) {
    xx = atpos[i];
    out->printf(">>  %d [%24.16e%24.16e%24.16e ] %d\n",
	      i, xx.v[0], xx.v[1], xx.v[2],
	      move[i]);
  }
}

void Speciesinfo::print_force(const Output *out, Everything &e) const
{
  vector3 ff;
  out->printf("Species %s, natoms = %d, forces follow:\n\n", 
	  name, natoms);
  for (int i=0; i < natoms; i++){
    if(e.ioninfo.coords_type == LATTICE_COORDS){
      ff = forces[i];
    } else if(e.ioninfo.coords_type == CARTESIAN_COORDS){
      ff = (~(inv3(e.lattice.latvec))) * forces[i];
    }else {
      die("Wrong CoordsType in Speciesinfo::print()\n");
    } 
    out->printf("%3d %18.10le %18.10le %18.10le %d\n",
	  i, ff.v[0], ff.v[1], ff.v[2], move[i]);
  }

  out->printf("\n");
}
