/*
    DFT++ is a density functional package developed by the research group
    of Professor Tomas Arias

    Copyright 1996-2003 Sohrab Ismail-Beigi

    This file is part of DFT++.

    DFT++ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DFT++ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DFT++; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please see the file CREDITS for a list of authors.

    For academic users, we request that publications using results obtained with
    this software reference

    "New algebraic formulation of density functional calculation," by Sohrab Ismail-Beigi
    and T.A. Arias, Computer Physics Communications 128:1-2, 1-45 (June 2000).

    and, if using the wavelet basis, further reference

    "Multiresolution analysis of electronic structure: semicardinal and wavelet bases,"
    T.A. Arias, Reviews of Modern Physics 71:1, 267-311 (January 1999).

    and 

    "Robust ab initio calculation of condensed matter: transparent convergence through
    semicardinal multiresolution analysis,'' I.P. Daykov, T.A. Arias, and
    Torkel D. Engeness, Physical Review Letters, 90:21, 216402 (May 2003).

    For your convenience, preprints of the above articles may be obtained from
    http://arXiv.org/abs/cond-mat/9909130, 9805262, and 0204411, respectively.
*/

#ifndef WL_COMPLEXCOLUMN_H
#define WL_COMPLEXCOLUMN_H

#define SPARSE 0
#define DENSE  1

class WL_ComplexColumn
{
 public:
  int fat;
  
  int length;
  ComplexArray data; // public. so that *Re and *Im can access it

  WL_BasisSpec *basis_spec;
  WL_Basis *basis;

  int representation;
  int embedding; // either SPARSE or DENSE

  // Constructors:
  WL_ComplexColumn();
  WL_ComplexColumn(WL_Basis *b, char s[10]="fat");
  WL_ComplexColumn(const WL_ComplexColumn &in, char s[10]="fat");

  ~WL_ComplexColumn();
  
  // Added to comply with PW_Column.
  // ipd: and probably will be used for double density scalar fields
  // ipd:  most likely not - vrfy and remove it if not needed
  void map(WL_ComplexColumn &c) const;

  char *getrepresentation() const;
  char *getembedding() const;

  void init_wavefunction(WL_Basis &b);
  void init_scalarfield(WL_Basis &b, int space);

  void randomize();
  void zero_out(){data.zero_out();};
  // Sets the column to all ones.
  void setones();
  // Sets the column to e(i*theta) on every point.
  void setmodones();

  void print();

  // Initializes a column with external data from binary files.
  void read(char *realfile, char *imagfile);
  // Initializes a column with real data from a text file, 
  // zeros imaginary part.
  void read_text(char *filename);

  // Reads and writes files in DFT++ format.
  void read(char *filename);
  void write(char *filename);

  void operator=(const WL_ComplexColumn &col);
  void operator-=(const WL_ComplexColumn &col);
  void operator+=(const WL_ComplexColumn &col);
  void operator*=(const real);
  void operator*=(const complex);
  void operator*=(const WL_ComplexColumn &col);
  complex operator^(const WL_ComplexColumn &);

  friend WL_ComplexColumn operator*(const real &r,
				    const WL_ComplexColumn &in)
    {
      WL_ComplexColumn out(in);
      out *= r;
      return out;
    }
  friend WL_ComplexColumn operator*(const complex &c,
				    const WL_ComplexColumn &in)
    {
      WL_ComplexColumn out(in);
      out *= c;
      return out;
    }
};
 
void GetRe(Grid *out, const WL_ComplexColumn &col);
void GetIm(Grid *out, const WL_ComplexColumn &col);
void PutRe(const Grid *in, WL_ComplexColumn &col);
void PutIm(const Grid *in, WL_ComplexColumn &col);

void add_scale_abs2(const complex &, const  WL_ComplexColumn &,
                    WL_ComplexColumn &);

WL_ComplexColumn abs2(WL_ComplexColumn &in);
WL_ComplexColumn pointwise_mult(const WL_ComplexColumn &a,
                                const WL_ComplexColumn &b);

// Does the dot product, assuming both Columns are exclusively real.
real dot(const WL_ComplexColumn &d1, const WL_ComplexColumn &d2);
 
#endif // WL_COMPLEXCOLUMN_H
