/*
    DFT++ is a density functional package developed by the research group
    of Professor Tomas Arias

    Copyright 1996-2003 Sohrab Ismail-Beigi

    This file is part of DFT++.

    DFT++ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DFT++ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DFT++; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please see the file CREDITS for a list of authors.

    For academic users, we request that publications using results obtained with
    this software reference

    "New algebraic formulation of density functional calculation," by Sohrab Ismail-Beigi
    and T.A. Arias, Computer Physics Communications 128:1-2, 1-45 (June 2000).

    and, if using the wavelet basis, further reference

    "Multiresolution analysis of electronic structure: semicardinal and wavelet bases,"
    T.A. Arias, Reviews of Modern Physics 71:1, 267-311 (January 1999).

    and 

    "Robust ab initio calculation of condensed matter: transparent convergence through
    semicardinal multiresolution analysis,'' I.P. Daykov, T.A. Arias, and
    Torkel D. Engeness, Physical Review Letters, 90:21, 216402 (May 2003).

    For your convenience, preprints of the above articles may be obtained from
    http://arXiv.org/abs/cond-mat/9909130, 9805262, and 0204411, respectively.
*/

#include "header.h"

void K(ColumnBundle &c, Everything &e)
{
  for(int i=0; i<c.my_ncols; i++)
    apply_K(*(c.col[i]),*(c.col[i]));
}

void apply_K(const WL_ComplexColumn &in, WL_ComplexColumn &out)
{
  Gridspec *gridspec;
  
  // choose the appropriate <sparse|dense> gridspec
  switch(in.embedding){
    case SPARSE:
      gridspec = in.basis_spec->gridspec;
      break;
    case DENSE:
      gridspec = in.basis_spec->gridspec_dense;
      break;
    default:
      die("apply_K: Unknown  input embedding %d\n",in.embedding);
  }

  Grid *grid = mkgridnew(gridspec,NULL);

  GetRe(grid,in);
  preconddiag(grid, 1);
  PutRe(grid,out);
  
  GetIm(grid,in);
  preconddiag(grid, 1);
  PutIm(grid,out);   

  killgridnew(grid);
}

/******************************************************/
/*Functions for preconddiag*/
void preconddiag(Grid *out, int recursive)
{
  double fact;
  int i, iend;
  double dx,dy,dz;
  int no_of_rows=out->dim.x*out->dim.y*out->dim.z;

  dx=out->scale.x;dy=out->scale.y;dz=out->scale.z;

  fact=3/( (dx*dy/dz)+(dy*dz/dx)+(dz*dx/dy))/(2.138499339301*2.);
//    fact=3/( (dx*dy/dz)+(dx*dy/dz)+(dx*dy/dz));

  // Avoid recursion:  all grids are "new" grids.  If we don't want a recursive
  // function, then restrict to the input grid.

  if (recursive)
    iend = out->nxyzwholegrid;
  else
    iend = out->nxyz;

  for (i=0;i<iend;i++)
    out->dat[i]*=fact;

//    if (recursive)
//      {
//        if (out->sibling)
//  	preconddiag(out->sibling,recursive);
//        if (out->child)
//  	preconddiag(out->child,recursive);
//      }
}
