/*
    DFT++ is a density functional package developed by the research group
    of Professor Tomas Arias

    Copyright 1996-2003 Sohrab Ismail-Beigi

    This file is part of DFT++.

    DFT++ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DFT++ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DFT++; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please see the file CREDITS for a list of authors.

    For academic users, we request that publications using results obtained with
    this software reference

    "New algebraic formulation of density functional calculation," by Sohrab Ismail-Beigi
    and T.A. Arias, Computer Physics Communications 128:1-2, 1-45 (June 2000).

    and, if using the wavelet basis, further reference

    "Multiresolution analysis of electronic structure: semicardinal and wavelet bases,"
    T.A. Arias, Reviews of Modern Physics 71:1, 267-311 (January 1999).

    and 

    "Robust ab initio calculation of condensed matter: transparent convergence through
    semicardinal multiresolution analysis,'' I.P. Daykov, T.A. Arias, and
    Torkel D. Engeness, Physical Review Letters, 90:21, 216402 (May 2003).

    For your convenience, preprints of the above articles may be obtained from
    http://arXiv.org/abs/cond-mat/9909130, 9805262, and 0204411, respectively.
*/

#include "command.h"

int ion_species_dep[] = {0};

extern char processing_error[DFT_LINE_LEN];

int
process_ion_species(char *input_line,Everything &e)
{
  char key[DFT_MSG_LEN];

  if (sscanf(input_line,"%*s %s",key)!=1)
    {
      sprintf(processing_error,
	      "ion-species must have a species name.  Input was:\n%s\n",
	      input_line);
      return 1;
    }
  for (int sp=0; sp < e.ioninfo.nspecies; sp++)
    if (MATCH(key,e.ioninfo.species[sp].name))
      {
	sprintf(processing_error,
		"Ion species %s is being multiply defined.\n",key);
	return 1;
      }

  real Z,mass;
  if (sscanf(input_line,"%*s %*s %lg %lg",
	     &Z,&mass) < 2)
    {
      sprintf(processing_error,
	      "ion-species must have at least 4 arguments.  Input was:\n%s\n",
	      input_line);
      return 1;
    }
  int &nsp = e.ioninfo.nspecies;
  nsp++;
  if (nsp==1)
    e.ioninfo.species = (Speciesinfo *)mymalloc(sizeof(Speciesinfo)*nsp,
						"species[]",
						"process_ion_species()");
  else
    e.ioninfo.species = (Speciesinfo *)myrealloc(e.ioninfo.species,
						 sizeof(Speciesinfo)*nsp,
						 "species[]",
						 "process_ion_species()");

  strcpy(e.ioninfo.species[nsp-1].name,key);
  e.ioninfo.species[nsp-1].Z = Z;
  e.ioninfo.species[nsp-1].mass = mass;

  return 0;
}

void
ion_species_print_status(Everything &e,Output *log)
{
  for (int sp=0; sp < e.ioninfo.nspecies; sp++)
    log->printf("ion-species %s %lg %lg",
		e.ioninfo.species[sp].name,
		e.ioninfo.species[sp].Z,
		e.ioninfo.species[sp].mass);
    log->printf("\n");
}

command *
setup_ion_species(command *next)
{
  command *cmnd = alloc_command();
  
  cmnd->number = ION_SPECIES;
  cmnd->name = "ion-species";
  cmnd->format = "ion-species <id> <Z> <mass>";
  cmnd->comments = "# Describe ion species:  type id, Z=nuclear charge, mass.\n";
  cmnd->dependencies = ion_species_dep;
  cmnd->process_command = process_ion_species;
  cmnd->print_status = ion_species_print_status;
  cmnd->allow_multiple = 1;
  cmnd->next_command = next;
  
  return cmnd;
}
