/*
    DFT++ is a density functional package developed by the research group
    of Professor Tomas Arias

    Copyright 1996-2003 Sohrab Ismail-Beigi

    This file is part of DFT++.

    DFT++ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DFT++ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DFT++; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please see the file CREDITS for a list of authors.

    For academic users, we request that publications using results obtained with
    this software reference

    "New algebraic formulation of density functional calculation," by Sohrab Ismail-Beigi
    and T.A. Arias, Computer Physics Communications 128:1-2, 1-45 (June 2000).

    and, if using the wavelet basis, further reference

    "Multiresolution analysis of electronic structure: semicardinal and wavelet bases,"
    T.A. Arias, Reviews of Modern Physics 71:1, 267-311 (January 1999).

    and 

    "Robust ab initio calculation of condensed matter: transparent convergence through
    semicardinal multiresolution analysis,'' I.P. Daykov, T.A. Arias, and
    Torkel D. Engeness, Physical Review Letters, 90:21, 216402 (May 2003).

    For your convenience, preprints of the above articles may be obtained from
    http://arXiv.org/abs/cond-mat/9909130, 9805262, and 0204411, respectively.
*/

#include "command.h"

int cntrl_algorithm_dep[] = {0};

#define N_ALGS 6

static char alg_names[N_ALGS][DFT_MSG_LEN] = 
  { "EOM", "PEOM", "SD", "PSD", "CG", "PCG" };

static int alg_numbers[N_ALGS] = 
  { DFT_EOM, DFT_PEOM, DFT_SD, DFT_PSD, DFT_CG, DFT_PCG };

extern char processing_error[DFT_LINE_LEN];

int
process_cntrl_algorithm(char *input_line,Everything &e)
{
  char key[DFT_MSG_LEN];
  sscanf(input_line,"%*s %s",key);

  int i;
  for (i=0; i < N_ALGS; i++)
    if (MATCH(key,alg_names[i]))
      {
	e.cntrl.algorithm = alg_numbers[i];
	return 0;
      }

  sprintf(processing_error,"Unknown minimization algorithm.\n");
  sprintf(processing_error,
	  "%s cntrl-algorithm must be one of the following:\n",
	  processing_error);
  for (i=0; i < N_ALGS; i++)
    if (i==0)
      sprintf(processing_error,"%s%s",processing_error,alg_names[i]);
    else
      sprintf(processing_error,"%s | %s",processing_error,alg_names[i]);
  sprintf(processing_error,"%s\n",processing_error);
  return 1;
}

void
cntrl_algorithm_set_default(Everything &e)
{
  // Default is PCG_nocos
  e.cntrl.algorithm = DFT_PCG;
}

void
cntrl_algorithm_print_status(Everything &e,Output *log)
{
  log->printf("cntrl-algorithm ");
  int i;
  for (i=0; i < N_ALGS; i++)
    if (e.cntrl.algorithm == alg_numbers[i])
      log->printf("%s\n",alg_names[i]);
}

command *
setup_cntrl_algorithm(command *next)
{
  static char format[DFT_LINE_LEN];
  command *cmnd = alloc_command();

  cmnd->number = CNTRL_ALGORITHM;
  cmnd->name = "cntrl-algorithm";
  sprintf(format,"%s",cmnd->name);
  int i;
  for (i=0; i < N_ALGS; i++)
    {
      if (i==0)
	sprintf(format,"%s %s",format,alg_names[i]);
      else
	sprintf(format,"%s | %s",format,alg_names[i]);
    }
  cmnd->format = format;
  cmnd->format = "cntrl-algorithm EOM | PEOM | SD | PSD | CG | PCG";
  cmnd->comments = "# Electronic minimization algorithm to use";
  cmnd->dependencies = cntrl_algorithm_dep;
  cmnd->has_default = 1;
  cmnd->process_command = process_cntrl_algorithm;
  cmnd->set_default = cntrl_algorithm_set_default;
  cmnd->print_status = cntrl_algorithm_print_status;
  cmnd->next_command = next;

  return cmnd;
}
