/*
    DFT++ is a density functional package developed by the research group
    of Professor Tomas Arias

    Copyright 1996-2003 Sohrab Ismail-Beigi

    This file is part of DFT++.

    DFT++ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DFT++ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DFT++; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please see the file CREDITS for a list of authors.

    For academic users, we request that publications using results obtained with
    this software reference

    "New algebraic formulation of density functional calculation," by Sohrab Ismail-Beigi
    and T.A. Arias, Computer Physics Communications 128:1-2, 1-45 (June 2000).

    and, if using the wavelet basis, further reference

    "Multiresolution analysis of electronic structure: semicardinal and wavelet bases,"
    T.A. Arias, Reviews of Modern Physics 71:1, 267-311 (January 1999).

    and 

    "Robust ab initio calculation of condensed matter: transparent convergence through
    semicardinal multiresolution analysis,'' I.P. Daykov, T.A. Arias, and
    Torkel D. Engeness, Physical Review Letters, 90:21, 216402 (May 2003).

    For your convenience, preprints of the above articles may be obtained from
    http://arXiv.org/abs/cond-mat/9909130, 9805262, and 0204411, respectively.
*/

#include "command.h"

int elec_ex_corr_dep[] = {SPINTYPE,0};

#define N_EXC 5

static char exc_names[N_EXC][DFT_MSG_LEN] = 
{ "lda", "lsd-teter", "gga-pw91", "gga-pbe", "sga-pbe" };
static int exc_numbers[N_EXC] = 
{ DFT_EXCORR_LDA, DFT_EXCORR_LSD_TETER, DFT_EXCORR_GGA_PW91, 
  DFT_EXCORR_GGA_PBE, DFT_EXCORR_SGA_PBE };

extern char processing_error[DFT_LINE_LEN];

int process_elec_ex_corr(char *input_line,Everything &e)
{
  char key[DFT_MSG_LEN];
  sscanf(input_line,"%*s %s",key);
  int i;
  for (i=0; i < N_EXC; i++)
    if (MATCH(key,exc_names[i]))
      {
	e.elecinfo.exc_opt = exc_numbers[i];
	if (e.elecinfo.spintype==NOSPIN &&
            ( exc_numbers[i]==DFT_EXCORR_LSD_TETER ||
	      exc_numbers[i]==DFT_EXCORR_SGA_PBE))
	  {
	    sprintf(processing_error,
		    "elec-ex-corr was given %s which is incompatible with spintype nospin\n",exc_names[i]);
	    return 1;
	  }
	else if (e.elecinfo.spintype==ZSPIN &&
	    (exc_numbers[i]==DFT_EXCORR_LDA ||
	     exc_numbers[i]==DFT_EXCORR_GGA_PW91 ||
	     exc_numbers[i]==DFT_EXCORR_GGA_PBE))
	  {
	    sprintf(processing_error,
		    "elec-ex-corr was given %s which is incompatible with spintype zspin\n",exc_names[i]);
	    return 1;
	  }
	return 0;
      }
  sprintf(processing_error,
	  "elec-ex-corr was given %s, but only the following are known:\n",
	  key);
  for (i=0; i < N_EXC; i++)
    if (i==0)
      sprintf(processing_error,"%s%s",processing_error,exc_names[i]);
    else
      sprintf(processing_error,"%s | %s",processing_error,exc_names[i]);
  sprintf(processing_error,"%s\n",processing_error);
  return 1;
}

void elec_ex_corr_set_default(Everything &e)
{
  // Default is LDA for no-spin, LSD-teter for z-spin
  if (e.elecinfo.spintype == NOSPIN)
    e.elecinfo.exc_opt = DFT_EXCORR_LDA;
  else
    e.elecinfo.exc_opt = DFT_EXCORR_LSD_TETER;
}

void elec_ex_corr_print_status(Everything &e,Output *log)
{
  for (int i=0; i < N_EXC; i++)
    if (e.elecinfo.exc_opt == exc_numbers[i])
      log->printf("elec-ex-corr %s\n",exc_names[i]);
}

command* setup_elec_ex_corr(command *next)
{
  command* cmnd = alloc_command();

  cmnd->number = ELEC_EX_CORR;
  cmnd->name = "elec-ex-corr";
  cmnd->format = "elec-ex-corr lda | lsd-teter | gga-pw91 | gga-pbe | sga-pbe";
  cmnd->comments = "# Specify the type of exchange-correlation functional";
  cmnd->dependencies = elec_ex_corr_dep;
  cmnd->has_default = 1;
  cmnd->process_command = process_elec_ex_corr;
  cmnd->set_default = elec_ex_corr_set_default;
  cmnd->print_status = elec_ex_corr_print_status;
  cmnd->next_command = next;

  return cmnd;
}
