/*
    DFT++ is a density functional package developed by the research group
    of Professor Tomas Arias

    Copyright 1996-2003 Sohrab Ismail-Beigi

    This file is part of DFT++.

    DFT++ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DFT++ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DFT++; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please see the file CREDITS for a list of authors.

    For academic users, we request that publications using results obtained with
    this software reference

    "New algebraic formulation of density functional calculation," by Sohrab Ismail-Beigi
    and T.A. Arias, Computer Physics Communications 128:1-2, 1-45 (June 2000).

    and, if using the wavelet basis, further reference

    "Multiresolution analysis of electronic structure: semicardinal and wavelet bases,"
    T.A. Arias, Reviews of Modern Physics 71:1, 267-311 (January 1999).

    and 

    "Robust ab initio calculation of condensed matter: transparent convergence through
    semicardinal multiresolution analysis,'' I.P. Daykov, T.A. Arias, and
    Torkel D. Engeness, Physical Review Letters, 90:21, 216402 (May 2003).

    For your convenience, preprints of the above articles may be obtained from
    http://arXiv.org/abs/cond-mat/9909130, 9805262, and 0204411, respectively.
*/

#include "command.h"

int ion_species_dep[] = {0};

extern char processing_error[DFT_LINE_LEN];

int process_ion_species(char *input_line,Everything &e)
{
#ifdef PLANEWAVES
  char key[DFT_MSG_LEN];
  if (sscanf(input_line,"%*s %s",key)!=1)
    {
      sprintf(processing_error,
	      "ion-species must have a species name.  Input was:\n%s\n",
	      input_line);
      return 1;
    }
  for (int sp=0; sp < e.ioninfo.nspecies; sp++)
    if (MATCH(key,e.ioninfo.species[sp].name))
      {
	sprintf(processing_error,
		"Ion species %s is being multiply defined.\n",key);
	return 1;
      }

  double Z, mass;
  char projp[DFT_MSG_LEN];
  char potfile[DFT_MSG_LEN],pulayfile[DFT_MSG_LEN];
  if (sscanf(input_line,"%*s %*s %lg %lg %s %s %s",
	     &Z,&mass,potfile,pulayfile,projp) < 4)
    {
      sprintf(processing_error,
	      "ion-species must have at least 5 arguments.  Input was:\n%s\n",
	      input_line);
      return 1;
    }
  int &nsp = e.ioninfo.nspecies;
  nsp++;
  if (nsp==1)
    e.ioninfo.species = (Speciesinfo *)mymalloc(sizeof(Speciesinfo)*nsp,
						"species[]",
						"process_ion_species()");
  else
    e.ioninfo.species = (Speciesinfo *)myrealloc(e.ioninfo.species,
						 sizeof(Speciesinfo)*nsp,
						 "species[]",
						 "process_ion_species()");

  e.ioninfo.species[nsp-1] = Speciesinfo(Z,key,0,0,mass) ;
  e.ioninfo.species[nsp-1].pot.potential_type = DFT_MIT_PSP;
  strcpy(e.ioninfo.species[nsp-1].pot.potfilename,potfile);
  strcpy(e.ioninfo.species[nsp-1].pot.pulayfilename,pulayfile);

  if (MATCH(projp, "projector")){
    e.ioninfo.species[nsp-1].pot.projectorp = 1;
    e.cntrl.projector_flag = TRUE;
  } else {
    e.ioninfo.species[nsp-1].pot.projectorp = 0;
  }


  if (sscanf(input_line,"%*s %*s %*g %*g %*s %*s %s",key) == 1)
    {
      if (MATCH(key, "fhi")) {
	e.ioninfo.species[nsp-1].pot.fhi_max_l     = 'd';
	e.ioninfo.species[nsp-1].pot.fhi_loc_l     = 'd';
	e.ioninfo.species[nsp-1].pot.fhi_ngrid_loc = 2004;
	e.ioninfo.species[nsp-1].pot.fhi_dq_loc    = 0.02;
	e.ioninfo.species[nsp-1].pot.fhi_ngrid_nl  =  501;
	e.ioninfo.species[nsp-1].pot.fhi_dq_nl     = 0.02;
	sscanf(input_line,"%*s %*s %*g %*g %*s %*s %*s %c %c %d %lg %d %lg",
	       &e.ioninfo.species[nsp-1].pot.fhi_max_l,
	       &e.ioninfo.species[nsp-1].pot.fhi_loc_l,
	       &e.ioninfo.species[nsp-1].pot.fhi_ngrid_loc, 
	       &e.ioninfo.species[nsp-1].pot.fhi_dq_loc, 
	       &e.ioninfo.species[nsp-1].pot.fhi_ngrid_nl, 
	       &e.ioninfo.species[nsp-1].pot.fhi_dq_nl);
	e.ioninfo.species[nsp-1].pot.potential_type = DFT_FHI_PSP;
      }
    }
#elif defined WAVELETS
  char key[DFT_MSG_LEN];
  if (sscanf(input_line,"%*s %s",key)!=1)
    {
      sprintf(processing_error,
	      "ion-species must have a species name.  Input was:\n%s\n",
	      input_line);
      return 1;
    }
  for (int sp=0; sp < e.ioninfo.nspecies; sp++)
    if (MATCH(key,e.ioninfo.species[sp].name))
      {
	sprintf(processing_error,
		"Ion species %s is being multiply defined.\n",key);
	return 1;
      }

  real Z,mass;
  char projp[DFT_MSG_LEN];
  char potfile[DFT_MSG_LEN],pulayfile[DFT_MSG_LEN];
  if (sscanf(input_line,"%*s %*s %lg %lg %s %s %s",
	     &Z,&mass,potfile,pulayfile,projp) < 4)
    {
      sprintf(processing_error,
	      "ion-species must have at least 5 arguments.  Input was:\n%s\n",
	      input_line);
      return 1;
    }
  int &nsp = e.ioninfo.nspecies;
  nsp++;
  if (nsp==1)
    e.ioninfo.species = (Speciesinfo *)mymalloc(sizeof(Speciesinfo)*nsp,
						"species[]",
						"process_ion_species()");
  else
    e.ioninfo.species = (Speciesinfo *)myrealloc(e.ioninfo.species,
						 sizeof(Speciesinfo)*nsp,
						 "species[]",
						 "process_ion_species()");

  strcpy(e.ioninfo.species[nsp-1].name,key);
  e.ioninfo.species[nsp-1].Z = Z;
  e.ioninfo.species[nsp-1].mass = mass;
#endif

  return 0;
}

void ion_species_print_status(Everything &e,Output *log)
{
#ifdef PLANEWAVES
  for (int sp=0; sp < e.ioninfo.nspecies; sp++) {
    log->printf("ion-species %s %lg %lg %s %s %s",
		e.ioninfo.species[sp].name,
		e.ioninfo.species[sp].Z,
		e.ioninfo.species[sp].mass,
		e.ioninfo.species[sp].pot.potfilename,
		e.ioninfo.species[sp].pot.pulayfilename,
		e.ioninfo.species[sp].pot.projectorp == 1 ? "projector":"");
    if (e.ioninfo.species[sp].pot.potential_type == DFT_FHI_PSP) {
      log->printf(" fhi %c %c %d %lg %d %lg",
		  e.ioninfo.species[sp].pot.fhi_max_l,
		  e.ioninfo.species[sp].pot.fhi_loc_l,
		  e.ioninfo.species[sp].pot.fhi_ngrid_loc,
		  e.ioninfo.species[sp].pot.fhi_dq_loc,
		  e.ioninfo.species[sp].pot.fhi_ngrid_nl,
		  e.ioninfo.species[sp].pot.fhi_dq_nl);
    }
    log->printf("\n");
  }
#endif
}

command* setup_ion_species(command *next)
{
  command* cmnd = alloc_command();

  cmnd->number = ION_SPECIES;
  cmnd->name = "ion-species";
  cmnd->format = "ion-species <id> <Z> <mass> <potfilename> <pulayfilename> [projector]\\\n# [fhi [<max-l-component> <local-l-component> <ngrid-loc> <dq-loc> <ngrid-nl> <dq-nl>]]";
  cmnd->comments = "# Describe ion species type id:  Z=nuclear charge, mass,\n# pseudopotential defined in file potfilename, and pulay info in pulayfilename.\n# pulayfile = none means no pulay correction to be done.\n#projector: pot file containts PAW projector information\n# (additional parameters to support FHI type potential:\n# fhi [<max-l-component> <local-l-component> <ngrid-loc> <dq-loc> \\\n# <ngrid-nl> <dq-nl>]";
  cmnd->dependencies = ion_species_dep;
  cmnd->process_command = process_ion_species;
  cmnd->print_status = ion_species_print_status;
  cmnd->allow_multiple = 1;
  cmnd->next_command = next;

  return cmnd;
}
