/*
    DFT++ is a density functional package developed by the research group
    of Professor Tomas Arias

    Copyright 1996-2003 Sohrab Ismail-Beigi

    This file is part of DFT++.

    DFT++ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DFT++ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DFT++; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please see the file CREDITS for a list of authors.

    For academic users, we request that publications using results obtained with
    this software reference

    "New algebraic formulation of density functional calculation," by Sohrab Ismail-Beigi
    and T.A. Arias, Computer Physics Communications 128:1-2, 1-45 (June 2000).

    and, if using the wavelet basis, further reference

    "Multiresolution analysis of electronic structure: semicardinal and wavelet bases,"
    T.A. Arias, Reviews of Modern Physics 71:1, 267-311 (January 1999).

    and 

    "Robust ab initio calculation of condensed matter: transparent convergence through
    semicardinal multiresolution analysis,'' I.P. Daykov, T.A. Arias, and
    Torkel D. Engeness, Physical Review Letters, 90:21, 216402 (May 2003).

    For your convenience, preprints of the above articles may be obtained from
    http://arXiv.org/abs/cond-mat/9909130, 9805262, and 0204411, respectively.
*/

/*
 * diag_matrix.C -- diagonal matrix class
 *
 */

/* $Id: diag_matrix.cpp,v 1.8.2.7 2003/05/29 18:54:19 ivan Exp $ */

#include "header.h"

/* default constructor */
diag_matrix::diag_matrix()
{
  n = 0; c = 0;
}

/* constructor with given size:  allocate the memory! */
diag_matrix::diag_matrix(int len)
{
  c = (scalar *)mymalloc(sizeof(scalar)*len,"c","diag_matrix(int)");
  n = len;
}

/* Copy constructor */
diag_matrix::diag_matrix(const diag_matrix &d1)
{
  int i;

  n = d1.n;
  c = (scalar *)mymalloc(sizeof(scalar)*n,"c","diag_matrix copy constructor");
  for (i=0; i < d1.n; i++)
    c[i] = d1.c[i];
}

/* desctructor */
diag_matrix::~diag_matrix()
{
  myfree(c);
}

/* Assignment:  nonstandard in that it returns void.  To make it standard,
 * replace void -> diag_matrix and uncomment the return *this; */
void diag_matrix::operator=(const diag_matrix &d1)
{
  int i;

  /* The sizes must agree */
  if (n != d1.n)
    die("in diag_matrix::operator=, n != d1.n\n");
  for (i=0; i < d1.n; i++)
    c[i] = d1.c[i];
  /* return *this; */
}

/* computes the trace of a matrix */

scalar
trace(const diag_matrix &dm)
{
  int i;
  scalar t=0.;
  for(i=0; i<dm.n; i++) t+=dm(i);

  return t;
}

/* Does what the constructor does as a member function */
void diag_matrix::init(int len)
{
  c = (scalar *)mymalloc(sizeof(scalar)*len,"c","diag_matrix.init");
  n = len;
}


void diag_matrix::print(Output *out)
{
  int i;

  for (i=0; i < n; i++)
    {
#if defined SCALAR_IS_COMPLEX
      out->printf("%lg%+lgi ",c[i].x,c[i].y);
#elif defined SCALAR_IS_REAL
      out->printf("%lg ",c[i]);
#else
#error scalar is neither real nor complex!
#endif
    }
  out->printf("\n");
}


void diag_matrix::print_real(Output *out)
{
  int i;

  for (i=0; i < n; i++)
#if defined SCALAR_IS_COMPLEX
    out->printf("%le ",c[i].x);
#else
#error scalar is not complex!
#endif
  
  out->printf("\n");
}



/* zeroes our the diagonal matrix */
void diag_matrix::zero_out(void)
{
  int i;
  for (i=0; i < n; i++)
    c[i] = (scalar)0.0;
}

// Linear Algebra
void
diag_matrix::operator*=(const complex &in)
{
  int i;
  for (i=0; i<n; i++)
    c[i]*=in;
}

diag_matrix operator*(const complex &c, const diag_matrix &in)
{
  diag_matrix out(in);
  out*=c;
  return out;
}
