/*
    DFT++ is a density functional package developed by the research group
    of Professor Tomas Arias

    Copyright 1996-2003 Sohrab Ismail-Beigi

    This file is part of DFT++.

    DFT++ is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DFT++ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DFT++; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please see the file CREDITS for a list of authors.

    For academic users, we request that publications using results obtained with
    this software reference

    "New algebraic formulation of density functional calculation," by Sohrab Ismail-Beigi
    and T.A. Arias, Computer Physics Communications 128:1-2, 1-45 (June 2000).

    and, if using the wavelet basis, further reference

    "Multiresolution analysis of electronic structure: semicardinal and wavelet bases,"
    T.A. Arias, Reviews of Modern Physics 71:1, 267-311 (January 1999).

    and 

    "Robust ab initio calculation of condensed matter: transparent convergence through
    semicardinal multiresolution analysis,'' I.P. Daykov, T.A. Arias, and
    Torkel D. Engeness, Physical Review Letters, 90:21, 216402 (May 2003).

    For your convenience, preprints of the above articles may be obtained from
    http://arXiv.org/abs/cond-mat/9909130, 9805262, and 0204411, respectively.
*/

#include "header.h"

void finitedifftest(Everything &e)
{
  dft_log("\nFinite difference test:\n");
  
  int nstates = e.elecinfo.nstates;
  int nbands = e.elecinfo.nbands;
  BlochState *states = e.elecvars.states;
  ColumnBundle **Y = e.elecvars.Y;
  CoeffSpaceScalarFieldColumn d_orig(e.elecvars.d);

  ColumnBundle **Ygrad = NULL;
  Matrix **Bgrad = NULL;
  Ygrad = alloc_ColumnBundle_array(nstates,states);
  Bgrad = alloc_Matrix_array(nstates,nbands,nbands);

  // Calculate and print the initial energies.
  calc_UVCn_d(e);
  d_orig = e.elecvars.d;
  calc_all_energies(e);
  e.energies.print(System::global_log);

  // Save the energies for future reference.
  dft_log("\nInitial energies:\n");
  Energies Eold = e.energies;

  // Get the initial gradient.
  calc_elecgrad_and_Hsub(Ygrad,Bgrad,e);
  
  real lderiv = 2.*dot(nstates,Ygrad,Ygrad);
  dft_log("\nLine derivative = %25.16le\n",lderiv);

  // Take steps in the Ygrad direction.
  for(real epsilon=1.; epsilon > 1.e-8; epsilon /= 10.)
    {
      // Take a step along Ygrad.
      scale_accumulate(nstates,epsilon,Ygrad,Y);
      
      // Recalculate energies.
      calc_UVCn_d_elec_dependent_energies(e);
      dft_log("\nepsilon = %1.12lf energies:\n",epsilon);
      e.energies.print(System::global_log);

      dft_log("\nOld total energy = %25.16le\n",Eold.Etot);
      dft_log("FD   Ratio: %25.16lf\n",
	      (e.energies.Etot-Eold.Etot)/(epsilon*lderiv));
      dft_log("FD sigfigs: %25.16lf\n",
	      1e-15*fabs(e.energies.Etot/(e.energies.Etot-Eold.Etot)));

      // Reset everything.
      scale_accumulate(nstates,-1.*epsilon,Ygrad,Y);
      calc_UVCn(e.elecinfo,e.elecvars,e.lattice,e.symm);
      e.elecvars.d = d_orig;      
    }
  
  // Free memory.
  free_ColumnBundle_array(nstates,Ygrad);
  free_Matrix_array(nstates,Bgrad);

}
